from __future__ import annotations

from collections.abc import Sequence
import itertools
from typing import Any, NamedTuple, Optional, Protocol, Union


# Vamos a usar una representación típica de la máquina de estados:
# una tabla state/event.
#
# Para declarar la table en el código escribiremos por cada columna:
#
# ```
# Current State = (
#     (input, input_arg, next_state),
#     ...
# )
# ```
#
# En la notación matemática convencional, la tabla viene siendo la
# función delta.

CommandName = str
StateName = str
DeltaArg = Tuple[
    Union[CommandName, Sequence[CommandName], Ellipsis],
    'domain.Domain',
    StateName
]

class _DeltaDescriptor:
    def __init__(self, args: list[DeltaArg]):
        self.args = args

    def __set_name__(self, owner, name):
        if not type(owner) == type:
            raise TypeError(f"delta() must be declared in class scope")

        owner._StateMachine__table[name] = self.args

        
def delta(*args) -> _DeltaDescriptor:
    return _DeltaDescriptor(args)


class StateMachine:
    __table = {}
    _cmds = {}

    def __init_subclass__(cls):
        # Calculamos los conjuntos de nombres de cmds y states de la tabla
        table_cmd_names = set()
        table_state_names = set()
        for cmds, _, next_st_name in itertools.chain(cls.__table.values()):
            if type(cmds) == list:
                table_cmd_names.update(cmds)
            elif cmds != ...:
                table_cmd_names.add(cmds)
            table_state_names.add(next_st_name)

        # Avisar de las funciones `do_xxx` que no se usan
        #
        # Lanzar una excepción si la tabla contiene un cmd para el que
        # no hay función `do_cmd`
        cls_cmds = { name.removeprefix(prefix): getattr(cls, name)
                 for name in dir(cls) if name.startswith(prefix) }
        valid_cmd_names = set(cls_cmds.keys())
        if ... not in table_cmd_names:
            unused_cmds = valid_cmd_names - table_cmd_names
            for cmd in unused_cmds:
                print(f"WARNING: the command {cmd} is not used in the state machine {cls.__name__}")
        invalid_cmds = table_cmd_names - valid_cmd_names
        invalid_cmds.discard(...)
        if len(invalid_cmds) > 0:
            fun_names = ", ".join(f"do_{cmd}" for cmd in invalid_cmds)
            raise ValueError(f"Missing {fun_names} in {cls.__name__}")
        cls._cmds = cls_cmds
        
        # TODO: no sabemos qué estados son inicial o final, cómo
        #   hacemos para avisar si hay un estado inalcanzable o del que
        #   no se puede salir (sobre todo si es por culpa de un typo)
        

# El objetivo no es "programar" una máquina de estados, sino definir
# un modelo abstracto del sistema que se está probando (_SUT_). Por
# eso la forma de definir la máquina de estados es un tanto particular.
#
# El caso que nos ocupa es un SUT donde el resultado de cada acción
# (_command_) depende del resultado de las acciones anteriores, es
# decir, del estado en que se encuentra el SUT. Al mismo tiempo el SUT
# es demasiado complejo como para poder escribir propiedades
# convencionales. Lo que hacemos es definir un modelo abstracto del
# SUT, suficientemente sencillo como para escribir propiedades
# relevantes sobre él. El modelo que elegimos se basa en una máquina
# de estados que define:
#
#   - Los estados en que puede estar el SUT.
#
#   - Las acciones (commands) que se pueden realizar sobre el SUT.
#
#   - Los cambios de estado que provoca cada acción.
#
# Una vez definido el modelo, el SUT pasa a ser una caja negra/gris.
# Sobre el modelo calcularemos los cambios de estado que provocan una
# serie de transiciones elegidas al azar. Al mismo tiempo, por cada
# transición lanzaremos la acción correspondiente sobre el SUT y
# comprobaremos las propiedades que nos garantizan que el
# comportamiento del SUT es coherente con el definido en el modelo.
#
# Por tanto, además de la máquina de estados, también necesitaremos:
#
#   - Funciones que lanzan en el SUT las acciones representadas en la
#     máquina de estados.
#
#
# Transition = (from_state, to_state, cmd_name, cmd_args, cmd_result)
#
#
# > \delta is the state-transition function: δ : S × Σ → S
# > {\displaystyle \delta :S\times \Sigma \rightarrow S}
#
# DeltaPairs = (StateName, Sequence ([command] | ..., Maybe Domain, StateName)
#
# Aquí hay que tener en cuenta que el orden es relevante en la
# secuencia de posibles transiciones. Igual que las cláusulas en erlang.

class Mi6STMachine(StateMachine):
    STOPPED = delta(
        ('Fundar', None, 'STARTED'),
        (..., None, 'STOPPED'), # No parece buena idea, pero completa el ejemplo
    )
    
    STARTED = delta(
        ('Recrutar', (domain.Axente(), domain.Destino()), 'STARTED'),
        ('Disolver', None, 'STOPPED'),
        (..., None, 'STARTED'),  
        (('Fundar', 'Asignar_Mision'), None, 'STARTED'),  # WARNING, el comodín lo subsume     
    )

    
    def __init__(self):
        self.axentes = {}

        
    # guardas estúpidas, ya pueden estar en las transiciones
    def guard_STARTED(self, from_state, cmd_args) -> bool:
        return from_state not in STARTED

    def guard_function(self, to_state, from_state, cmd_args) -> bool:
        return from_state not in STOPPED
    

    def do_Fundar(self, cmd_args) -> Any:
        ...

    def do_Recrutrar(self, cmd_args) -> Any:
        ...

    def do_Asignar_Mision(self, cmd_args) -> Any:
        ...

    def do_Consultar_Estado(self, cmd_args) -> Any:
        ...

    def do_Disolver(self, cmd_args) -> Any:
        ...


    def predicate_Consultar_Estado(self, transition: Transition) -> bool:
        return self.axentes[transition.cmd_args] == transition.cmd_result


# STMRun(state_machine: StateMachine, initial_state, sut_factory: Callable[[], Any])
#
# El dominio STMRun es el conjunto de todas las secuencias de transiciones que se pueden
# dar en la máquina de estados, partiendo del estado indiciado.
#
# Cada secuencia de transisiones se representa como un `run`, que es
# un iterable de dicha secuencia.
#
# run = Iterable[Transistion]
#
# Obviously cada transición depende de las anteriores (eso que llaman
# estado).  Las transiciones se aplican a medida que se itera sobre
# ellas (de forma perezosa). Por cada transición podemos comprobar los
# predicados que consideremos oportunos.

@forall(run= STMRun(Echo, Echo.STOPPED, echo_factory))
def test_todo_esta_en_su_sitio(run):
    for transition in run:
        # Comprobamos los predicados definidos en la StateMachine
        # Si queremos comprobar otra cosa, pues adelante
        Echo.check(transition)
    # También podríamos hacer
    Echo.check(run)




















class _Name:
    _name = None

    def __str__(self):
        return self._name
    
    def __repr__(self):
        return f"{self._name} = {self.__class__.__name__}()"

    def __set_name__(self, owner, name):
        if not type(owner) == type:
            raise TypeError(f"{self.__class__.__name__} must be declared in class scope")
        self._name = f"{owner.__name__}.{name}"


class StateName(_Name)
    def __contains__(self, other):
        if type(other) == StateName:
            return other == self
        elif isinstance(other, State):
            return other._state_name == self
        else:
            return False

    @property
    def name(self) -> StateName:
        return self

    def replace(self, name) -> State:
        raise ValueError("Cannot change name in StateName")
    
        
class State(Protocol):
    name: StateName
    def replace(self, **kwargs) -> State:
        ...

    
class CommandName(_Name):
    def __call__(self, event: Union[Domain, Any]= None) -> Command:
        return Command(self, event)


class Command:
    def __init__(self, name: CommandName, event: Union[Domain, Any]= None):
        self.name: CommandName = name
        self._event: Union[Domain, Any] = event

    def events(self) -> Iterator:
        if not is_domain(self.event):
            yield self.event
        else:
            yield from self._event()




# Ejemplos

class EchoSTMachine(StateMachine):
    STOPPED = StateName()
    STARTED = StateName()

    Start = CommandName()
    Stop  = ComnnandName()
    Print = CommandName()

    transitions = {
        STOPPED: [
            (Start, None, STARTED)
        ],
        STARTED: [
            (Print, domain.String(), STARTED),
            (Stop, None, STOPPED)
        ],
    }

    def do_Print(self, event):
        ...

    def do_Start(self):
        ...

    def do_Stop(self):
        ...
        

class Mi6STMachine(StateMachine):
    STARTED = StateName()
    STOPPED = StateName()

    Fundar = CommandName()
    Recrutar = CommandName()
    Asignar_Mision = CommandName()
    Consultar_Estado = CommandName()
    Disolver = CommandName()

    transitions = {
        STOPPED: [
            (Fundar, None, STARTED),
            (..., None, STOPPED), # No parece buena idea, pero completa el ejemplo
        ],
        STARTED: [
            (Recrutar, (domain.Axente(), domain.Destino()), STARTED),
            (Disolver, None, STOPPED),
            (..., None, STARTED),  
            (Fundar | Asignar_Mision), None, STARTED),  # WARNING, el comodín lo subsume     
        ]


# El objetivo no es "programar" una máquina de estados, sino definir
# un modelo abstracto del sistema que se está probando (_SUT_). Por
# eso la forma de definir la máquina de estados es un tanto particular.
#
# El caso que nos ocupa es un SUT donde el resultado de cada acción
# (_command_) depende del resultado de las acciones anteriores, es
# decir, del estado en que se encuentra el SUT. Al mismo tiempo el SUT
# es demasiado complejo como para poder escribir propiedades
# convencionales. Lo que hacemos es definir un modelo abstracto del
# SUT, suficientemente sencillo como para escribir propiedades
# relevantes sobre él. El modelo que elegimos se basa en una máquina
# de estados que define:
#
#   - Los estados en que puede estar el SUT.
#
#   - Las acciones (commands) que se pueden realizar sobre el SUT.
#
#   - Los cambios de estado que provoca cada acción.
#
# Una vez definido el modelo, el SUT pasa a ser una caja negra/gris.
# Sobre el modelo calcularemos los cambios de estado que provocan una
# serie de transiciones elegidas al azar. Al mismo tiempo, por cada
# transición lanzaremos la acción correspondiente sobre el SUT y
# comprobaremos las propiedades que nos garantizan que el
# comportamiento del SUT es coherente con el definido en el modelo.
#
# Por tanto, además de la máquina de estados, también necesitaremos:
#
#   - Funciones que lanzan en el SUT las acciones representadas en la
#     máquina de estados.
#
#
# Transition = (from_state, to_state, cmd_name, cmd_args, cmd_result)
#
#
# > \delta is the state-transition function: δ : S × Σ → S
# > {\displaystyle \delta :S\times \Sigma \rightarrow S}
#
# DeltaPairs = (StateName, Sequence ([command] | ..., Maybe Domain, StateName)
#
# Aquí hay que tener en cuenta que el orden es relevante en la
# secuencia de posibles transiciones. Igual que las cláusulas en erlang.

class Mi6STMachine(StateMachine):
    STOPPED = delta(
        ('Fundar', None, 'STARTED'),
        (..., None, 'STOPPED'), # No parece buena idea, pero completa el ejemplo
    )
    
    STARTED = delta(
        ('Recrutar', (domain.Axente(), domain.Destino()), 'STARTED'),
        ('Disolver', None, 'STOPPED'),
        (..., None, 'STARTED'),  
        (('Fundar', 'Asignar_Mision'), None, 'STARTED'),  # WARNING, el comodín lo subsume     
    )

    
    def __init__(self):
        self.axentes = {}

        
    # guardas estúpidas, ya pueden estar en las transiciones
    def guard_STARTED(self, from_state, cmd_args) -> bool:
        return from_state not in STARTED

    def guard_function(self, to_state, from_state, cmd_args) -> bool:
        return from_state not in STOPPED
    

    def do_Fundar(self, cmd_args) -> Any:
        ...

    def do_Recrutrar(self, cmd_args) -> Any:
        ...

    def do_Asignar_Mision(self, cmd_args) -> Any:
        ...

    def do_Consultar_Estado(self, cmd_args) -> Any:
        ...

    def do_Disolver(self, cmd_args) -> Any:
        ...


    def predicate_Consultar_Estado(self, transition: Transition) -> bool:
        return self.axentes[transition.cmd_args] == transition.cmd_result


# STMRun(state_machine: StateMachine, initial_state, sut_factory: Callable[[], Any])
#
# El dominio STMRun es el conjunto de todas las secuencias de transiciones que se pueden
# dar en la máquina de estados, partiendo del estado indiciado.
#
# Cada secuencia de transisiones se representa como un `run`, que es
# un iterable de dicha secuencia.
#
# run = Iterable[Transistion]
#
# Obviously cada transición depende de las anteriores (eso que llaman
# estado).  Las transiciones se aplican a medida que se itera sobre
# ellas (de forma perezosa). Por cada transición podemos comprobar los
# predicados que consideremos oportunos.

@forall(run= STMRun(Echo, Echo.STOPPED, echo_factory))
def test_todo_esta_en_su_sitio(run):
    for transition in run:
        # Comprobamos los predicados definidos en la StateMachine
        # Si queremos comprobar otra cosa, pues adelante
        Echo.check(transition)
    # También podríamos hacer
    Echo.check(run)




################################################################################

class State(NamedTuple):
    name: StateName


def state(st: Union[StateName, State]) -> State:
    if isinstance(st, StateName):
        return State(st)
    else:
        return st


def _normalize(transition_list):
    return [ transition if type(transition) == tuple else (None, transition)
        for transition in transition_list
    ]


class StateMachine:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        print(cls, kwargs)
        cls.transitions = { st: _normalize(transition_list)
                            for st, transition_list in cls.transitions.items() }
        print(cls.transitions)

    def __setattr__(self, name, value):
        raise TypeError("Instances of StateMachine are inmutable")

    def __str__(self):
        return f"StateMachine:{self.__class__.__name__}"

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def get_transition(self, state: State) -> Optional[Transition]:
        transitions = self.transitions[state.name]
        for transition in _random.sample(transitions, k= len(transitions)):
            event_domain, next_state = transition
            guard = getattr(self, f"guard_{next_state}", None)
            if guard is None:
                return transition
            elif not is_domain(event_domain):
                event = event_domain
                if guard(event):
                    return Transition(event, next_state)
            else:
                dom = event_domain()
                # Si la transición tiene asociado un dominio, intentamos n veces
                # encontrar una muestra que cumpla la guarda
                for _, event in zip(range(100), dom):
                    if guard(event):
                        return Transition(event, next_state)
        return None

    def transition_to(self, from_state: State, transition: Transition) -> State:
        next_state = from_state._replace(name= transition.state_name)
        before = getattr(self, f"before_{str(state.name)}", None)
        if before is not None:
            next_state = before(next_state, transition.event)
        after = getattr(self, f"after_{str(state.name)}", None)
        if before is not None:
            next_state = after(next_state, transition.event)
        return next_state

    def do_sut_transition_to(self, sut: Any, from_state: State, transition: Transition) -> Any:
        command = getattr(self, f"do_command_to_{str(state.name)}", None)
        if command is None:
            raise RuntimeError(f"Cannot transition sut to {state.name}")
        return command(sut, from_state, transition.event)
    
    
class Transition(NamedTuple):
    event: Optional[Union[Domain, Any]]
    state_name: StateName
        

class FooState(State):
    field_1: int= 0
    field_2: str= ""

    def __init__(self, state_name: StateName):
        super().__init__(state_name)
        
    
class Foo(StateMachine):
    CREATED = StateName()
    PARTIAL = StateName()
    COMPLETE = StateName()

    transitions = {
        CREATED: [ ('domain.Int()', PARTIAL) ],
        PARTIAL: [ COMPLETE ],
    }

    def guard_function(self, state, event) -> bool:
        return state in self.CREATED or state in self.PARTIAL

    def before_transition(self, state, event) -> Foo:
        return data

    def after_transition(self, state, event) -> Foo:
        return data

    def do_transition(self, state, event):
        sut.do_something()


print("Creando")
stm = Foo()
print(stm)
#stm.field_1 = 7
#stm.transitions = {}
#stm.CREATED = 0
print()
st = FooState(stm.CREATED)
print(st in stm.CREATED)
print(st in stm.PARTIAL)
print(stm.CREATED in stm.CREATED)
stm2 = Foo()
print(stm2.CREATED in stm.CREATED)


class Echo(StateMachine):
    STOPPED = StateName()
    STARTED = StateName()

    START = CommandName()
    PRINT = CommandName()
    STOP = CommandName()
    
    transitions = {
        STARTED: [ (PRINT, domain.String()), (STOP, None) ],
        STOPPED: [ (START, None) ]
    }

"""
property "o servidor de echo imprime todo o que se lle manda", [:verb
       │ ose] do
   9   │     forall cmds in commands(__MODULE__) do
  10   │       trap_exit do
  11   │         kill_echo_if_alive()
  12   │         {history, state, result} = run_commands(__MODULE__, cmds)
  13   │ 
  14   │         (result == :ok)
  15   │         |> when_fail(
  16   │           IO.puts("""
  17   │           History: #{inspect(history, pretty: true)}
  18   │           State: #{inspect(state, pretty: true)}
  19   │           Result: #{inspect(result, pretty: true)}
  20   │           """)
  21   │         )
  22   │         |> aggregate(command_names(cmds))
  23   │       end
  24   │     end
  25   │   end
"""

# Estaría bien si pudiesemos poner el sut en cualquier estado arbitrario
#@forall(state= State(STMFoo))
#@forall(transition= lambda state: Transition(STMFoo, state))
#def test_o_servidor_rula(state, transition):
#    next_state = STMFoo.do_transition(state, transition, sut)
#    return STMFoo.check_sut_state(sut, next_state)


def echo_factory():
    pass


@forall(run= STMRun(Echo, Echo.STOPPED, echo_factory))
def test_o_sut_rula(run):
    for step in run:
        if step.is_error():
            return False
    return True


class STMRun(Domain):
    def __init__(self,
                 stm: StateMachine,
                 initial_state: Union[StateName, State],
                 sut_factory: Callable[[], Any]):
        self.stm = stm
        self.initial_state = state(initial_state)
        self.sut_factory = sut_factory

    def __iter__(self) -> Iterator:
        while True:
            yield self._one_run()

    def _one_run(self) -> Iterator[Step]:
        state = self.initial_state
        stm = self.stm
        sut = self.sut_factory()
        while True:
            transition = stm.get_transition(state)
            next_state = stm.transition_to(state, transition)
            sut = stm.do_sut_transition_to(sut, state, transition)

            # comparar sut y next_state
            # yield resultado

    def __str__(self):
        return f"STMRun({self.stm}, {self.initial_state}, {self.sut_factory})"



#!/usr/bin/env python3

from __future__ import annotations

from datetime import timedelta
import importlib
import inspect
import os
from pathlib import Path
import sys
import time
from typing import Any
import unittest

from rich import box, pretty
from rich.abc import RichRenderable
from rich.columns import Columns
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, BarColumn, MofNCompleteColumn, SpinnerColumn, TaskProgressColumn, TimeElapsedColumn, TimeRemainingColumn, TextColumn
from rich.rule import Rule
from rich import status
from rich.syntax import Syntax
from rich.text import Text
from rich.traceback import Traceback

from pypbt import PyPBT, domains
from pypbt.quantifiers import CounterExample, is_qcproperty, PredicateError


IGNORE = (
    '__pycache__',
)

# TODO
Candidate = object


def code_panel(
        fun: Callable,
        title: str= "",
        title_align: str= 'left',
        truncate: bool= True,
) -> Panel:
    try:
        source_lines, line_no = inspect.getsourcelines(fun)
    except OSError:
        source_lines = None
    if source_lines is not None:
        if not truncate:
            source = "".join(source_lines)
        else:
            i = 0
            for line in source_lines:
                if line.strip().startswith("def "):
                    break
                i += 1
            if len(source_lines) < i + 4:
                source = "".join(source_lines)
            else:
                source = "".join(source_lines[:i+4]) + "···"

        # TODO: Truncate after n lines of predicate
        s = Syntax(
            code= source.strip(),
            lexer= 'python',
            line_numbers= True,
            start_line= line_no
        )
    else:
        s = Syntax(code= str(fun), lexer= 'python')
    return Panel(s, box= box.ROUNDED, title= title, title_align= title_align)


def env_panel(env: dict, title:str= "") -> Panel:
    s = "\n".join(f"{k} = {pretty.traverse(v)}" for k, v in env.items())
    return Panel(s, box= box.ROUNDED, expand= False, title= title)


def traceback_from_exception(exc: Exception) -> Traceback:
    return Traceback.from_exception(
        exc_type= type(exc),
        exc_value= exc,
        traceback= exc.__traceback__,
        show_locals= True,
    )


def spinner_progress(spinner:str= 'pong') -> Progress:
    return  Progress(
        TextColumn("[progress.description]{task.description}"),
        SpinnerColumn(spinner),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        transient= True,
    )


class ReporterMixin:
    def _report_summary(self, *lines: Sequence[str], console: Console, title: str) -> None:
        console.print(Panel("\n".join(lines), title= title))


# --------------------------------------------------------------------------------------
# Unittest
# --------------------------------------------------------------------------------------
loader = unittest.TestLoader()


class UnittestResult(unittest.TestResult):
    def __init__(self, console: Console, status: Status):
        super().__init__()
        self.console = console
        self.status = status
        self.tests_run = 0
        self.n_success = 0

    def get_description(self, test):
        doc_first_line = test.shortDescription()
        if doc_first_line:
            return f"{test}\n{doc_first_line}"
        else:
            return test._testMethodName
        
    def startTest(self, test):
        self.tests_run += 1
        self.status.update(status= f"  {self.get_description(test)} ...")

    def addSubTest(self, test, subtest, err):
        super().addSubTest(test, subtest, err)
        if err is None:
            self.n_success += 1
        else:
            if issubclass(err[0], subtest.failureException):
                exctype, value, tb = err
                # Sólo queremos el último frame. Es donde está el assert que falló.
                while tb.tb_next is not None:
                    tb = tb.tb_next
                tb = Traceback.from_exception(exctype, value, tb, show_locals= True)
                self.console.print(tb)
                self.console.print(f"{self.get_description(test)} [red]FAILED[/red]")
            else:
                exctype, value, tb = err
                tb = Traceback.from_exception(exctype, value, tb, show_locals= True)
                self.console.print(tb)
                self.console.print(f"{self.get_description(test)} [red]ERROR[/red]")
   
    def addSuccess(self, test):
        self.n_success += 1
        self.console.print(f"{self.get_description(test)} [green]PASSED[/green]")

    def addError(self, test, err):
        super().addError(test, err)
        exctype, value, tb = err
        tb = Traceback.from_exception(exctype, value, tb, show_locals= True)
        self.console.print(tb)
        self.console.print(f"{self.get_description(test)} [red]ERROR[/red]")


    def addFailure(self, test, err):
        super().addFailure(test, err)
        exctype, value, tb = err
        # Sólo queremos el último frame. Es donde está el assert que falló.
        while tb.tb_next is not None:
            tb = tb.tb_next
        tb = Traceback.from_exception(exctype, value, tb, show_locals= True)
        self.console.print(tb)
        self.console.print(f"{self.get_description(test)} [red]FAILED[/red]")

    def addSkip(self, test, reason):
        super().addSkip(test, reason)
        self.console.print(
            f"{self.get_description(test)}"
            f" [yellow]SKIPPED [dim]\"{reason}\"[/dim][/yellow]"
        )

    def addExpectedFailure(self, test, err):
        super().addExpectedFailure(test, err)
        exctype, value, tb = err
        self.console.print(
            f"{self.get_description(test)}"
            f" [yellow]Expected failure {exctype.__name__}(\"{value}\")[/yellow]"
        )
        
    def addUnexpectedSuccess(self, test):
        super().addUnexpectedSuccess(test)
        self.console.print(
            f"{self.get_description(test)}"
            f" [red]Unexpected success[/red]"
        )

        
class UnittestRunner(ReporterMixin):
    def __init__(self):
        self.tests_run = 0
        self.n_success = 0
        self.n_errors = 0
        self.n_failures = 0
        self.n_skipped = 0
        self.n_expected_failures = 0
        self.n_unexpected_successes = 0
        self.elapsed_total = 0
                
    def wants_to_iter_dir(self, dir: Path) -> bool:
        return True

    def wants_to_run_file(self, file: Path) -> bool:
        return file.suffix == ".py"

    def will_do(self, test: Any) -> bool:
        return isinstance(test, type) and issubclass(test, unittest.TestCase)

    def do_test(self, test: Any, console: Console) -> None:
        console.rule(title= f"[b]{test.__name__}[/b]", align= 'left')
        tests = loader.loadTestsFromTestCase(test)
        with console.status(f"[b]{test.__name__}[/b]") as status:
            result = UnittestResult(console, status)
            start_time = time.perf_counter()
            tests(result)
            stop_time = time.perf_counter()
            e_time = stop_time - start_time
        self.elapsed_total += e_time
        self.tests_run += result.tests_run
        self.n_success += result.n_success
        self.n_errors += len(result.errors)
        self.n_failures += len(result.failures)
        self.n_skipped += len(result.skipped)
        self.n_expected_failures += len(result.expectedFailures)
        self.n_unexpected_successes += len(result.unexpectedSuccesses)
        console.rule(title= "")
        console.line()

    def report_summary(self, console: Console) -> None:
        if self.tests_run == 0:
            return
        delta = timedelta(seconds= self.elapsed_total)
        self._report_summary(
            f"tests run: {self.tests_run} [progress.elapsed]({delta})[/]",
            f"success: {self.n_success}",
            f"failures: {self.n_failures}",
            *([f"errors: {self.n_errors}"] if self.n_errors > 0 else []),
            *([f"skipped: {self.n_skipped}"] if self.n_skipped > 0 else []),
            *([f"expected failures: {self.n_expected_failures}"] if self.n_expected_failures > 0 else []),
            *([f"unexpected successes: {self.n_unexpected_successes}"] if self.n_unexpected_successes > 0 else []),
            title= f"[b]unittest[/b]",
            console= console,
        )
        

# --------------------------------------------------------------------------------------
# Hypothesis
# --------------------------------------------------------------------------------------
class HypothesisRunner(ReporterMixin):
    def __init__(self):
        self.n_properties = 0
        self.n_falsifying_examples = 0
        self.n_errors = 0
        self.elapsed_total = 0

    def wants_to_iter_dir(self, dir: Path) -> bool:
        return True

    def wants_to_run_file(self, file: Path) -> bool:
        return file.suffix == ".py"

    def will_do(self, test: Any) -> bool:
        return getattr(test, 'is_hypothesis_test', False)

    def do_test(self, test: Any, console: Console) -> None:
        self.n_properties += 1
        console.print(code_panel(test))
        from contextlib import redirect_stdout
        import io
        capture = io.StringIO()
        try:
            start_time = time.perf_counter()
            with redirect_stdout(capture):
                test()
            stop_time = time.perf_counter()
            console.print("[green]PASSED[/green]")
        except AssertionError:
            stop_time = time.perf_counter()
            self.n_falsifying_examples += 1
            console.print_exception()
            console.print(Panel(capture.getvalue(), title= "", expand= False))
        except:
            stop_time = time.perf_counter()
            self.n_errors += 1
            console.print_exception()
        e_time = stop_time - start_time
        self.elapsed_total += e_time
        console.line()
        
    def report_summary(self, console: Console) -> None:
        if self.n_properties == 0:
            return
        delta = timedelta(seconds= self.elapsed_total)
        self._report_summary(
            f"properties checked: {self.n_properties} [progress.elapsed]({delta})[/]",
            f"total number of falsifying examples found: {self.n_falsifying_examples}",
            *([f"total number of buggy properties: {self.n_errors}"] if self.n_errors > 0 else []),
            title= "[b]Hypothesis[/b]",
            console= console,
        )

        
# --------------------------------------------------------------------------------------
# PyPBT
# --------------------------------------------------------------------------------------
class PyPbtRunner(ReporterMixin):
    def __init__(self):
        self.n_properties = 0
        self.n_samples = 0
        self.n_counterexamples = 0
        self.n_errors = 0
        self.elapsed_total = 0
        
    def wants_to_iter_dir(self, dir: Path) -> bool:
        return True

    def wants_to_run_file(self, file: Path) -> bool:
        return file.suffix == ".py"
    
    def will_do(self, test: Any) -> bool:
        return is_qcproperty(test)

    def do_test(self, test: Any, console: Console) -> None:
        prop = test
        console.print(code_panel(prop.get_predicate()))
        self.n_properties += 1
        with spinner_progress() as progress:
            task = progress.add_task("Testing...", total= None)
            for i, result in enumerate(prop(env= {}), start= 1):
                if result:
                    progress.advance(task)
                    self.n_samples += 1
                elif isinstance(result, CounterExample):
                    progress.stop()
                    self.n_counterexamples += 1
                    console.print(env_panel(result.env, title= " counterexample "))
                    n_tests = f"{i} tests" if i>1 else "1 test"
                    msg = f"[red]FAILED[/red] after {n_tests}"
                    break
                elif isinstance(result, PredicateError):
                    progress.stop()
                    self.n_errors += 1
                    console.print(traceback_from_exception(result.exc))
                    console.print(env_panel(result.env, title= " counterexample "))
                    n_tests = f"{i} tests" if i>1 else "1 test"
                    msg = f"[red]FAILED[/red] after {n_tests}"
                    break
                else:
                    raise RuntimeError("fUnkown {result=}")
            else:
                progress.stop()
                n_tests = f"{i} tests" if i>1 else "1 test"
                msg = f"[green]PASSED[/green] {n_tests}"
            elapsed = progress.tasks[task].elapsed
            delta = timedelta(seconds= elapsed)
            console.print(f"{msg} [progress.elapsed]({delta})[/]")
            self.elapsed_total += elapsed
            console.line(2)

    def report_summary(self, console: Console) -> None:
        if self.n_properties == 0:
            return
        delta = timedelta(seconds= self.elapsed_total)
        self._report_summary(
            f"properties checked: {self.n_properties} [progress.elapsed]({delta})[/]",
            f"total number of samples checked: {self.n_samples}",
            f"total number of counterexamples found: {self.n_counterexamples}",
            *([f"number of errors: {self.n_errors}"] if self.n_errors > 0 else []),
            title= f"[b]{PyPBT}[/b]",
            console= console,
        )


# --------------------------------------------------------------------------------------
# EBT
# --------------------------------------------------------------------------------------
TEST_ATTR = '__test_title__'


def test(title: str) -> Callable[[Callable],Callable]:
    def decorator(fun: Callable) -> Callable:
        setattr(fun, TEST_ATTR, title)
        return fun
    return decorator

    
class KissEBTRunner(ReporterMixin):
    def __init__(self):
        self.tests_run = 0
        self.n_success = 0
        self.n_failures = 0
        self.n_errors = 0
        self.elapsed_total = 0

    def want_to_iter_dir(self, dir: Path) -> bool:
        return True

    def wants_to_run_file(self, file: Path) -> bool:
        return file.suffix == ".py"

    def will_do(self, test: Any) -> bool:
        return hasattr(test, TEST_ATTR)

    def do_test(self, test: Any, console: Console) -> None:
        title = getattr(test, TEST_ATTR)
        with console.status(f"{title} ..."):
            start_time = time.perf_counter()
            try:
                test()
            except Exception as e:
                stop_time = time.perf_counter()
                exception = e
            else:
                stop_time = time.perf_counter()
                exception = None
        self.tests_run += 1
        e_time = stop_time - start_time
        self.elapsed_total += e_time
        delta = timedelta(seconds= e_time)
        if exception is None:
            self.n_success += 1
            console.print(f"[b]{title}[/b] [green]PASSED[/green] [progress.elapsed]({delta})[/]")
        elif isinstance(exception, AssertionError):
            self.n_failures += 1
            console.print(traceback_from_exception(exception))
            console.print(f"[b]{title}[/b] [red]FAILED[/red] [progress.elapsed]({delta})[/]")
        else:
            self.n_errors += 1
            console.print_exception()
            console.print(
                f"[b]{title}[/b] [red]ERROR \"Unexpected exception\"[/red]"
                f" [progress.elapsed]({delta})[/]")

        
    def report_summary(self, console: Console) -> None:
        if self.tests_run == 0:
            return
        delta = timedelta(seconds= self.elapsed_total)
        self._report_summary(
            f"tests run: {self.tests_run} [progress.elapsed]({delta})[/]",
            f"success: {self.n_success}",
            f"failures: {self.n_failures}",
            *([f"errors: {self.n_errors}"] if self.n_errors > 0 else []),
            title= f"[b]EBT[/b]",
            console= console,
        )


# --------------------------------------------------------------------------------------
# Runner
# --------------------------------------------------------------------------------------

def collect_objs_from_file(file: Path, console: Console) -> Iterator:
    parent = str(file.parent)
    module_name = file.stem
    if parent not in sys.path:
        sys.path.append(parent)
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        console.print_exception()
        module = None
    if module is not None:
        for name in dir(module):
            obj = getattr(module, name)
            yield obj
        

def collect_objs(
        path: Path,
        candidates: tuple[Candidate],
        console: Console
) -> Iterator[tuple[Any, tuple[Candidate]]]:
    if path.name in IGNORE:
        return
    if path.is_dir():
        dir_name = f"{path}{os.sep}"
        candidates= tuple(candidate for candidate in candidates
                          if candidate.wants_to_iter_dir(path))
        if len(candidates) ==  0:
            console.print(Rule(title= f" Skipping (no runners): {dir_name} ", characters= "="))
        else:
            console.print(Rule(title= f" [b]Entering dir: {dir_name}[/b] ", characters= "="))
            console.line()
            for child in path.iterdir():
                yield from collect_objs(
                    path= child,
                    candidates= candidates,
                    console= console
                )
            console.print(Rule(title= f" Leaving: {dir_name} ", characters= "-"))
            console.line()
                
    elif path.is_file():
        candidates= tuple(candidate for candidate in candidates
                          if candidate.wants_to_run_file(path))
        if len(candidates) ==  0:
            console.print(Rule(title= f" Skipping (no runners): {path} ", characters= "="))
        else:
            console.print(Rule(title= f" [b]Collecting: {path}[/b] ", characters= "="))
            console.line()
            for obj in collect_objs_from_file(file= path, console= console):
                yield obj, candidates
            console.print(Rule(title= f" Finished: {path} ", characters= "-"))
            console.line()
                
    elif not path.exists():
        # TODO: Better message
        print(f"ERROR: File not found: {path}")
        sys.exit(1)
        
    else:
        # TODO: Better message
        raise RuntimeError(f"{path} is not regular file nor directory")
                
    
def main():
    if len(sys.argv) < 2:
        cmd = Path(sys.argv[0]).name
        print(f"USAGE: {cmd} <props_files|dirs>")
        sys.exit(1)

    CANDIDATES = (
        PyPbtRunner(),
        HypothesisRunner(),
        UnittestRunner(),
        KissEBTRunner(),
    )
    seed = domains.seed
    console = Console()
    for arg in sys.argv[1:]:
        path = Path(arg)
        for obj, candidates in collect_objs(
                path= path,
                candidates= CANDIDATES,
                console= console):
            candidate = next((candidate for candidate in candidates
                              if candidate.will_do(obj)), None)
            if candidate is not None:
                candidate.do_test(obj, console)
    console.print(
        Panel(
            f"started with seed: {seed}",
            title= ""
        )
    )
    for candidate in CANDIDATES:
        candidate.report_summary(console)
        
            
if __name__ == '__main__':
    main()



from __future__ import annotations

from itertools import islice
import textwrap
from typing import Callable, get_args, Literal, NamedTuple, Union

from .domains import domain_expr, Domain, DomainCoercible


DEFAULT_N_SAMPLES = 100


# Property = ForAll Property
#          | Exists Property
#          | Predicate
#
# Críticas a esta definición de `Property`:
#
#   - `Property = Predicate` es un caso muy extremo de propiedad.
#
#   - Nada impide que el programado construya mal una propiedad. En
#     concreto puede terner variables libres, i.e., en nuestro caso,
#     sin cuantificar.  Sin embargo el conjunto de errores que puede
#     cometer un programador es infinito. No podemos evitarlos
#     todos. Tampoco es algo que pudiese evitar la mayoría de sistemas
#     de tipos.
#
# TODO: No está reflejado el hecho de que una propiedad pueda tener
# variables, libres o ligadas.
#
# TODO: Las variables libres en un predicado pueden no ser tal cosa si
# incorporamos conceptos como las `fixtures` de `pytest`. En este caso
# son variables que no están ligadas en ningún cuantificador, pero
# están ligadas en otro contexto más global (y compartido por varias
# propiedades).


# :( `property` is a built-in function in python
class QCProperty: pass


def is_qcproperty(x):
    return isinstance(x, QCProperty)


def qcproperty(arg) -> QCProperty:
    if is_qcproperty(arg):
        return arg
    else:
        return Predicate(arg)


# Entorno. Guarda las ligaduras de variables.
VarName = str
Env = dict[VarName, DomainCoercible]


# --------------------------------------------------------------------------------------
# Domain LambdaExpr
# --------------------------------------------------------------------------------------
# Abstractions containing free variables that will evaluate to a Domain.
# The term should be a DomainCoercible expression.
#
# Thus, actually, `Callable` means `lambda`.
DomainLambdaExpr = Callable[[...], Domain]

def is_lambda_expr(arg: QArg) -> bool:
    return getattr(arg, '__name__', None) == '<lambda>'


BOUND_VARS_ATTR = '__domain_lambda_expr_bound_vars__'


def domain_lambda_expr(expr: Callable) -> DomainLambdaExpr:
    # El cálculo de las bound_vars lo hacemos aquí en lugar de hacerlo
    # en `reduce_expr` porque esa función se llama por cada sample del
    # quantificador "padre":
    #
    # @forall(xs= domain.List(domain.Int(), min_len= 4, max_len= 4))
    # @forall(x= lambda xs: domain_expr(xs, is_exhaustible= True))
    signature = inspect.signature(expr)
    bound_vars = list(signature.parameters.keys())
    if len(bound_vars) == 0:
        raise TypeError(f"no bound variables in {expr}")
    setattr(expr, BOUND_VARS_ATTR, bound_vars)
    return expr


def reduce_expr(expr: DomainLambdaExpr, env: Env) -> Domain:
    bound_vars = getattr(expr, BOUND_VARS_ATTR, None)
    if bound_vars is None:
        return expr
    # A la hora de construir kwargs ignoramos la variables que no
    # están en env para que el error salte al llamar a la función
    kwargs = { k: env[k] for k in bound_vars if k in env }
    return domain_expr(expr(**kwargs))


# --------------------------------------------------------------------------------------
# Syntax sugar
# --------------------------------------------------------------------------------------
"""Values binded to the quantified variables.

It's not very sound, because a `DomainLambadTerm` is also a
`DomainCoercible`. As any object is coercible.
"""
QArg = Union[DomainLambdaExpr, DomainCoercible]


def desugar_var_value(arg: QArg) -> Domain:
    if is_lambda_expr(arg):
        return domain_lambda_expr(arg)
    else:
        return domain_expr(arg)


#---------------------------------------------------------------------------
# 
#---------------------------------------------------------------------------
class CounterExample(NamedTuple):
    env: Env

    def __bool__(self) -> bool:
        return False

    
class PredicateError(NamedTuple):
    exc: Exception
    env: Env

    def __bool__(self) -> bool:
        return False
    
    
CheckResult = Union[
    Literal[True],
    CounterExample,
    PredicateError,
]


#---------------------------------------------------------------------------
# Predicados
#---------------------------------------------------------------------------
# Implementación (función python) de un predicado
PredicateFun = Callable[..., bool]


# Adaptador de predicado a QC.
class Predicate(QCProperty):
    def __init__(self, pred: PredicateFun):
        self.pred = pred
        
    def __call__(self, /, env: Env) -> Iterator[CheckResult]:
        try:
            result = self.pred(**env) or CounterExample(env)
        except Exception as e:
            result = PredicateError(exc= e, env= env)
        yield result

    def __str__(self):
        return str(self.pred.__name__)

    def get_predicate(self) -> PredicateFun:
        return self.pred
    

#---------------------------------------------------------------------------
# Cuantificadores
#---------------------------------------------------------------------------
# Podríamos implementar los cuantificadores como funciones (cierres).
# Los implementamos así para incluir otras operaciones interesantes
# como p.e. `__str__`

# Todas las propiedades se "ejecutan" en un `Environment` que contiene
# las variables ligadas por cada uno de los cuantificadores.


class ForAll(QCProperty):
    def __init__(self,
                 quantified_var: VarName,
                 domain_obj: Domain,
                 qcproperty: QCProperty,
                 n_samples: int):
        self.quantified_var = quantified_var
        self.domain_obj = domain_obj
        self.qcproperty = qcproperty
        self.n_samples = n_samples

    def __call__(self, /, env: Env) -> Iterator[CheckResult]:
        quantified_var = self.quantified_var
        if quantified_var in env:
            raise TypeError(f"Variable {quantified_var} is shadowed in {self}")

        domain_obj = reduce_expr(self.domain_obj, env)
        prop = self.qcproperty

        # Si el dominio está marcado como finito recorremos el dominio entero
        if domain_obj.is_exhaustible:
            domain_samples = domain_obj.exhaustible
        else:
            domain_samples = islice(domain_obj, self.n_samples)
            
        for sample in domain_samples:
            yield from prop(env= {**env, quantified_var: sample})

    def __str__(self) -> str:
        return (
            f"ForAll {self.quantified_var} in {self.domain_obj}\n"
            f"{textwrap.indent(str(self.qcproperty), '  ')}"
        )

    def get_predicate(self) -> str:
        return self.qcproperty.get_predicate()

        
class Exists(QCProperty):
    def __init__(self,
                 quantified_var: VarName,
                 domain_obj: Domain,
                 qcproperty: QCProperty):
        if not isinstance(qcproperty, Predicate):
            # TODO: Podríamos comprobar que todos los cuantificadores dentro
            #       de este Exists son exhaustivos ?
            #       Seguramente es más sencillo añadir un parámetro opcional
            #       a la función `qc` que indique si tiene que ser exhaustive
            raise TypeError(f"This tool cannot check Exists on another quantifier")
        
        self.quantified_var = quantified_var
        self.domain_obj = domain_obj
        self.qcproperty = qcproperty

        
    def __call__(self, /, env: Env) -> Iterator[CheckResult]:
        quantified_var = self.quantified_var
        if quantified_var in env:
            raise TypeError(f"Variable {quantified_var} is shadowed in {self}")

        domain_obj = reduce_expr(self.domain_obj, env)
        prop = self.qcproperty

        if not domain_obj.is_exhaustible:
            raise TypeError(f"It's not possible to check existence "
                            f"in non exhaustive domain: {domain_obj}")
        
        for sample in domain_obj.exhaustible:
            env = {**env, quantified_var: sample}
            for result in prop(env= env):
                if result:
                    yield True
                    break
                elif isinstance(result, PredicateError):
                    yield result
                    break
            else:
                yield CounterExample(env)

    def get_predicate(self) -> str:
        return self.qcproperty.get_predicate()
    
    def __str__(self):
        return (
            f"Exists {self.quantified_var} in {self.domain_obj}\n"
            f"{textwrap.indent(str(self.qcproperty), '  ')}"
        )
        
    
#---------------------------------------------------------------------------
# Decorators
#---------------------------------------------------------------------------
def forall(n_samples: int= DEFAULT_N_SAMPLES, **binds):
    """Decorates a predicate funcion or another decorator with a forall quantifier.

    Parameters
    ----------
    n_samples : int
        The number of samples to check.
    **binds
        The quantified variables. The number of quantified
        variables must be 1. In order to quantify more than
        one variable, you should write each variable in its own
        `@forall`.


    Returns
    -------
    ForAll
        An object implementing the property as a callable.
    """
    if len(binds) != 1:
        # TODO: ¿ queremos implementar forall(x= ..., y= ...) como
        #         forall(x,y= ...,...) ?
        #       Es decir forall(t= Tuple(...,...) x,y= t
        raise TypeError(f"Must bind just one variable, but {len(binds)} binded")
    var, obj = next(iter(binds.items()))
    
    def factory(arg):
        return ForAll(
            quantified_var= var,
            domain_obj= desugar_var_value(obj),
            qcproperty= qcproperty(arg),
            n_samples= n_samples
        )

    return factory


def exists(**binds):
    """Decorates a predicate funcion or another decorator with an existencial quantifier.

    Parameters
    ----------
    **binds
        The quantified variables. The number of quantified
        variables must be 1. In order to quantify more than
        one variable, you should write each variable in its own
        `@exists`.


    Returns
    -------
    Exists
        An object implementing the property as a callable.

    !!! warning
        Every variable must be quantified over an exhaustible domain.

    """
    if len(binds) != 1:
        raise TypeError(f"Must bind just one variable, but {len(binds)} binded")
    var, obj = next(iter(binds.items()))
    
    def factory(arg):
        return Exists(
            quantified_var= var,
            domain_obj= desugar_var_value(obj),
            qcproperty= qcproperty(arg)
        )

    return factory

